from deltalake.azure import AzureDeltaReader
from deltalake.local import LocalDeltaReader


class DeltaReader(AzureDeltaReader, LocalDeltaReader):
    """
    Main entry point for interacting with delta tables.
    Instantiating this class will only try to find the underlying files for the table.
    No data is read until you call the `to_pyarrow` or `to_pandas` functions.
    Parameters:
    ----------
    path: (string)
        Path to the delta table. Class will try to derive the storage location from the path.
        Currently only supports local and Azure.

    credential: (string, azure.identity class)
        credential to authenticate against storage.
        For local file system this can be omitted, but for Azure it should be a
        SAS-token, Access Key or `azure.identity` class

    Returns:
    -------
    dr : (DeltaReader)
        Delta reader that has parsed the log files
    """

    def __new__(self, path, credential=None):
        if "abfss://" in path:
            return AzureDeltaReader(path, credential)
        elif "https://" in path:
            return AzureDeltaReader(path, credential)
        elif "s3://" in path:
            raise NotImplementedError("S3 not supported yet")
        elif "file://" in path:
            return LocalDeltaReader(path)
        elif "://" not in path:
            return LocalDeltaReader(path)
        raise NotImplementedError(
            f"Could not find supported storage integration for {path}"
        )



import json
import re
from copy import deepcopy
from io import BytesIO

import pyarrow
import pyarrow.parquet as pq
from azure.core.exceptions import ResourceNotFoundError
from azure.storage.blob import ContainerClient

from deltalake.base import BaseDeltaReader


class AzureDeltaReader(BaseDeltaReader):
    def __init__(self, path, credential=None):
        super(AzureDeltaReader, self).__init__(path, credential)

    def _parse_path(self):
        if "dfs.core.windows.net" in self.path:
            raise ValueError(
                "use blob.core.windows.net instead of dfs.core.windows.net"
            )

        if "https://" in self.path:
            m = re.search("https://(.+?).blob.core.windows.net/(.+?)/(.+)", self.path)
            self.account_name = m.group(1)
            self.container_name = m.group(2)
            self.folder = m.group(3)

        elif "abfss://" in self.path:
            m = re.search("abfss://(.+?)@(.+?).blob.core.windows.net/(.+)", self.path)
            self.container_name = m.group(1)
            self.account_name = m.group(2)
            self.folder = m.group(3)

    def _authenticate(self):
        self.container_client = ContainerClient(
            account_url=f"https://{self.account_name}.blob.core.windows.net",
            container_name=self.container_name,
            credential=self.credential,
        )

    def _download_blob(self, file):
        bc = self.container_client.get_blob_client(blob=file)
        return bc.download_blob()

    def _is_delta_table(self):
        blobs = self.container_client.list_blobs(
            name_starts_with=f"{self.folder}/_delta_log/{0:020}.json"
        )
        try:
            blobs.next()
            return True

        except StopIteration:
            return False

    def _apply_from_checkpoint(self, checkpoint_version: int):

        # reset file set, and checkpoint version
        self.files = set()
        self.checkpoint = checkpoint_version

        if self.checkpoint == 0:
            return

        # read latest checkpoint
        checkpoint_blob = self._download_blob(
            f"{self.folder}/_delta_log/{self.checkpoint:020}.checkpoint.parquet"
        )
        # Read as stream to minimize overhead
        with BytesIO() as checkpoint_stream:
            checkpoint_blob.download_to_stream(checkpoint_stream)
            # Convert stream to pandas table using pyarrow
            checkpoint = pq.read_table(checkpoint_stream).to_pandas()

            for i, row in checkpoint.iterrows():
                added_file = row["add"]["path"] if row["add"] else None
                if added_file:
                    self.files.add(added_file)

    def _apply_partial_logs(self, version: int):
        # Checkpoints are created every 10 transactions,
        # so we need to find all log files with version
        # up to 9 higher than checkpoint.
        # Effectively, this means that we can just create a
        # wild card for the first decimal of the checkpoint version

        log_files = self.container_client.list_blobs(
            name_starts_with=f"{self.folder}/_delta_log/{self.checkpoint//10:019}"
        )
        # sort the log files, so we are sure we get the correct order
        log_files = sorted(log_files, key=lambda log: log.name)
        for log_file in log_files:
            # skip checkpoint files
            if ".json" in log_file.name:

                # Get version from log name
                log_version = re.findall(r"(\d{20})", log_file.name)[0]
                self.version = int(log_version)

                # Download log file
                log = self._download_blob(log_file).readall()
                for line in log.split():
                    meta_data = json.loads(line)
                    # Log contains other stuff, but we are only
                    # interested in the add or remove entries
                    if "add" in meta_data.keys():
                        self.files.add(meta_data["add"]["path"])
                    if "remove" in meta_data.keys():
                        remove_file = meta_data["remove"]["path"]
                        # To handle 0 checkpoints, we might read the log file with
                        # same version as checkpoint. this means that we try to
                        # remove a file that belongs to an ealier version,
                        # which we don't have in the list
                        if remove_file in self.files:
                            self.files.remove(remove_file)
                # Stop if we have reatched the desired version
                if self.version == version:
                    break

    def _as_newest_version(self):
        # Try to get the latest checkpoint info
        try:
            # get latest checkpoint version
            checkpoint_info_blob = self._download_blob(
                file=f"{self.folder}/_delta_log/_last_checkpoint"
            )
            checkpoint_info = json.loads(checkpoint_info_blob.readall())
            self._apply_from_checkpoint(checkpoint_info["version"])

        except ResourceNotFoundError:
            pass

        # apply remaining versions. This can be a maximum of 9 versions.
        # we will just break when we don't find any newer logs
        self._apply_partial_logs(version=self.checkpoint + 9)

    def to_pyarrow(self, columns=None):
        tables = []
        for file in self.files:

            with BytesIO() as stream:
                # File contains full path, which we cannot use with the container client.
                # If you create a blob client for each blob, it will introduce a lot of overhead
                blob = self._download_blob(f"{self.folder}/{file}")

                blob.download_to_stream(stream)
                tables.append(pq.read_table(stream, columns=columns))

        return pyarrow.concat_tables(tables)

    def as_version(self, version: int):
        nearest_checkpoint = version // 10

        deltaReader = deepcopy(self)
        deltaReader._apply_from_checkpoint(nearest_checkpoint)
        deltaReader._apply_partial_logs(version=version)

        return deltaReader



import json
from copy import deepcopy
from os import path

import pyarrow.parquet as pq

from deltalake.base import BaseDeltaReader


class LocalDeltaReader(BaseDeltaReader):
    def __init__(self, path: str):
        super(LocalDeltaReader, self).__init__(path)

    def _parse_path(self):
        # just remove file prefix, since we don't need this
        if "file://" in self.path:
            self.path = self.path.replace("file://", "")

    def _is_delta_table(self):
        return path.exists(self.log_path)

    def _apply_from_checkpoint(self, checkpoint_version: int):

        # reset file set, and checkpoint version
        self.files = set()
        self.checkpoint = checkpoint_version

        if self.checkpoint == 0:
            return

        checkpoint = pq.read_table(
            f"{self.log_path}/{self.checkpoint:020}.checkpoint.parquet"
        ).to_pandas()

        for i, row in checkpoint.iterrows():
            added_file = row["add"]["path"] if row["add"] else None
            if added_file:
                self.files.add(f"{self.path}/{added_file}")

    def _apply_partial_logs(self, version: int):
        for i in range(version - self.checkpoint + 1):
            try:
                with open(f"{self.log_path}/{self.checkpoint+i:020}.json", "r") as f:
                    self.version = self.checkpoint + i
                    for line in f:
                        meta_data = json.loads(line)
                        # Log contains other stuff, but we are only
                        # interested in the add or remove entries
                        if "add" in meta_data.keys():
                            self.files.add(f"{self.path}/{meta_data['add']['path']}")
                        if "remove" in meta_data.keys():
                            remove_file = f"{self.path}/{meta_data['remove']['path']}"
                            # To handle 0 checkpoints, we might read the log file with
                            # same version as checkpoint. this means that we try to
                            # remove a file that belongs to an ealier version,
                            # which we don't have in the list
                            if remove_file in self.files:
                                self.files.remove(remove_file)

            # If the file isn't found it should be because we have reatched
            # the newest log file in last iteration
            except FileNotFoundError:
                break

    def _as_newest_version(self):
        # Check if we have any checkpoints before reading any files
        if path.exists(f"{self.log_path}/_last_checkpoint"):
            with open(f"{self.log_path}/_last_checkpoint", "r") as f:
                checkpoint_info = json.load(f)
                checkpoint = checkpoint_info["version"]
                # apply versions from checkpoint
                self._apply_from_checkpoint(checkpoint)

        # apply remaining versions. This can be a maximum of 9 versions.
        # we will just break when we don't find any newer logs
        self._apply_partial_logs(version=self.checkpoint + 9)

    def to_pyarrow(self, columns=None):
        return pq.ParquetDataset(list(self.files)).read_pandas(columns=columns)

    def as_version(self, version: int):
        nearest_checkpoint = version // 10

        deltaReader = deepcopy(self)
        deltaReader._apply_from_checkpoint(nearest_checkpoint)
        deltaReader._apply_partial_logs(version=version)

        return deltaReader



import pandas as pd


class BaseDeltaReader:
    def __init__(self, path, credential=None):
        self.path = path
        self.credential = credential
        self.version = 0
        self.checkpoint = 0
        self.files = set()

        self._parse_path()
        self._authenticate()

        self.log_path = f"{self.path}/_delta_log"

        if not self._is_delta_table():
            raise ValueError(f"No delta table found in {self.path}")

        self._as_newest_version()

    def _parse_path(self):
        pass

    def _authenticate(self):
        pass

    def _is_delta_table(self):
        pass

    def _as_newest_version(self):
        pass

    def as_version(self, version):
        pass

    def to_pyarrow(self, columns=None):
        pass

    def to_pandas(self, columns=None) -> pd.DataFrame:
        """Reads the Delta table as a pandas dataframe

        Parameters:
        ----------
        columns : ([string], optional)
            Only read a subset of columns from the table. Defaults to None.

        Returns:
        -------
        df : (pd.Dataframe)
            Content of Delta table as pandas dataframe
        """
        return self.to_pyarrow(columns=columns).to_pandas()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deltalake']

package_data = \
{'': ['*']}

install_requires = \
['azure-storage-blob>=12.6.0,<13.0.0', 'pyarrow>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'delta-lake-reader',
    'version': '0.1.1',
    'description': 'Lightweight wrapper for reading Delta tables without Spark',
    'long_description': '![Build package](https://github.com/jeppe742/DeltaLakeReader/workflows/Build%20python%20package/badge.svg)\n# Delta Lake Reader\nThe [Delta](https://github.com/delta-io/delta) format, developed by Databricks, is often used to build data lakes.\n\nWhile it tries to solve many issues with data lakes, one of the downsides is that delta tables rely on Spark to read the data. If you only need to read a small table, this can introduce a lot of unnecessary overhead.\n\nThis package tries to fix this, by providing a lightweight python wrapper around the delta file format.\n\n\n\n# Usage\nPackage currently only support local file system, and azure blob storage, but should be easily extended to AWS and GCP in the future.\nThe main entry point should be the `DeltaReader` class. This will try to derrive the underlying file system, based on the input URL.\n\nWhen the class is instantiated, it will try to parse the transaction log files, to find the files in the newest table version. It will, however, not read any data before you run the `to_pyarrow` or `to_pandas` functions.\n## Local file system\n\n```python\nfrom deltalake import DeltaReader\n\n# native file path\ntable_path = "somepath/mytable"\n# Get table as pyarrow table\ndf = DeltaReader(table_path).to_pyarrow()\n# Get table as pandas dataframe\ndf = DeltaReader(table_path).to_pandas()\n\n\n# file url\ntable_path = "file://somepath/mytable"\ndf = DeltaReader(table_path).to_pandas()\n```\n## Azure\nThe Azure integration is based on the [Azure python SDK](https://docs.microsoft.com/en-us/azure/developer/python/azure-sdk-overview). The `credential` used to authenticate against the storage account, can be either a SAS token, Access Keys or one of the `azure.identity` classes ([read more](https://docs.microsoft.com/en-us/python/api/azure-identity/azure.identity?view=azure-python)).\n\nThe input path can either be the https or abfss protocol (will be converted to https under the hood). Note that the current implementation doesn\'t support the `dfs.core.windows.net` api. But you should simply be able to replace dfs with blob.\n```python\nfrom deltalake import DeltaReader\n\ncredential = "..." #SAS-token, Access keys or an azure.identity class\n\n#abfss\ntable_url = "abfss://mycontainer@mystorage.blob.core.windows.net/mytable"\ndf = DeltaReader(table_url, credential).to_pandas()\n\n#https\ntable_url = "https://mystorage.blob.core.windows.net/mycontainer/mytable"\ndf = DeltaReader(table_url, credential).to_pandas()\n```\n\n## Time travel\nOne of the features of the Delta format, is the ability to do timetravel.\n\nThis can be done using the `as_version` property. Note that this currenly only support specific version, and not timestamp.\n```python\nfrom deltalake import DeltaReader\n\ntable_url = "https://mystorage.blob.core.windows.net/mycontainer/mytable"\ncredential = "..."\ndf = DeltaReader(table_url, credential).as_version(5).to_pandas()\n```\n\n\n## Disclaimer\nDatabricks recently announced a stand alone reader for Delta tables in a [blogpost](https://databricks.com/blog/2020/12/22/natively-query-your-delta-lake-with-scala-java-and-python.html)\nThe python bindings mentioned, however, requires you to install the rust library which might sound scary for a python developer.\n\n# Read more\n[Delta transaction log](https://databricks.com/blog/2019/08/21/diving-into-delta-lake-unpacking-the-transaction-log.html)\n',
    'author': 'Jeppe Johan Waarkjær Olsen',
    'author_email': 'jeppe742@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jeppe742/DeltaLakeReader',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
